import collections
import string
import numpy as np
import logging
from decimal import Decimal
from collections import Iterable
from openpyxl.styles.numbers import *

ERR_NUM = '#NUM!'
ERR_NULL = '#NULL!'
ERR_NA = '#N/A'
ERR_VALUE = '#VALUE!'
ERR_REF = '#REF!'
XLFN = '_xlfn.'
ERR_NAME = '#NAME?'
ERR_DIV = '#DIV/0!'

ERROR_VALUES = [ERR_DIV, ERR_NA, ERR_NAME, ERR_NULL, ERR_NUM, ERR_REF, ERR_VALUE]

LOG_FINE = 8

logger = logging.getLogger(__name__)


"""
Remove any trailing zeros as fractional part
Example: 5.500 -> 5.5
Ref: https://stackoverflow.com/questions/11227620/drop-trailing-zeros-from-decimal
"""
def normalize_fraction(d):
    if not isinstance(d, Decimal):
        d = Decimal(d)
    normalized = d.normalize()
    sign, digit, exponent = normalized.as_tuple()
    return normalized if exponent <= 0 else normalized.quantize(1)
"""
It will convert the list of rows to a dictionary such that the keys correspond 
to the first column in the list of row and the values to the column indicated through the 
argument col index.
"""
def to_dictionary(row_list, col_index_num):
    return {row[0]: row[col_index_num] for row in row_list}

def create_address(rng, sheet=None):
    if is_collection(rng):
        address = ''
        for rg in rng:
            if len(address) > 0:
                address += ';'
            address += create_address_single(rg, sheet)
        return address
    else:
        return create_address_single(rng, sheet)

def create_address_single(rng, sheet=None):
    address = rng.replace('$', '')
    if sheet is not None and address.find('!') < 0:
        address = sheet + '!' + address
    return address

def safe_str(value):
    return str(value) if value is not None else ''

def get_range(spreadsheet, cell_address):
    return spreadsheet.cellmap[cell_address].address()

def evaluate_spreadsheet(spreadsheet, sheet, col_nr, row_nr):
    return spreadsheet.evaluate(sheet + '!' + col_nr + str(row_nr))

def to_values(spreadsheet, rng, skip_none=True, is_valid_function=None):
    result = []
    def process(col_nr, row_nr, value):
        if not skip_none or value is not None and is_valid_function(value):
            result.append(value)
        return True
    iterate_cells(spreadsheet, rng, process)
    return result

# Change item from Uppercase to Lowercase
def lower_case_item(item):
    if isinstance(item, Iterable):
        item = [str(entry).lower() for entry in item]
        return item
    else:
        return str(item).lower()

"""
rng can be a single range string or a tuple of ranges.
"""
def iterate_cells(spreadsheet, rng, process_fn):
    if isinstance(rng, list):
        for rg in rng:
            iterate_cells_single(spreadsheet, rg, process_fn)
    else:
        iterate_cells_single(spreadsheet, rng, process_fn)

"""
process_fn will receive the absolute col and row, both in numbers. If returned, false, iteration will stop.
"""
def iterate_cells_single(spreadsheet, rng, process_fn):
    sheet, start, end = split_range(rng)
    sh, start_col, start_row = split_address(start)
    sh, end_col, end_row = split_address(end)
    start_col_nr, end_col_nr = col2num(start_col), col2num(end_col)
    start_row_nr, end_row_nr = int(start_row), int(end_row)
    for row_nr in range(start_row_nr, end_row_nr + 1):
        for col_nr in range(start_col_nr, end_col_nr + 1):
            try:
                if not process_fn(col_nr, row_nr, evaluate_spreadsheet(spreadsheet, sheet, num2col(col_nr), row_nr)):
                    return
            except:
                raise Exception("Error iterating: col_nr: [%s], row_nr: [%s], address: [%s]" % (col_nr, row_nr, rng))

def rankdata(a, method='average'):
    """
    Assign ranks to data, dealing with ties appropriately.
    Ranks begin at 1.  The `method` argument controls how ranks are assigned
    to equal values.  See [1]_ for further discussion of ranking methods.
    Parameters
    ----------
    a : array_like
        The array of values to be ranked.  The array is first flattened.
    method : str, optional
        The method used to assign ranks to tied elements.
        The options are 'average', 'min', 'max', 'dense' and 'ordinal'.
        'average':
            The average of the ranks that would have been assigned to
            all the tied values is assigned to each value.
        'min':
            The minimum of the ranks that would have been assigned to all
            the tied values is assigned to each value.  (This is also
            referred to as "competition" ranking.)
        'max':
            The maximum of the ranks that would have been assigned to all
            the tied values is assigned to each value.
        'dense':
            Like 'min', but the rank of the next highest element is assigned
            the rank immediately after those assigned to the tied elements.
        'ordinal':
            All values are given a distinct rank, corresponding to the order
            that the values occur in `a`.
        The default is 'average'.
    Returns
    -------
    ranks : ndarray
         An array of length equal to the size of `a`, containing rank
         scores.
    References
    ----------
    .. [1] "Ranking", http://en.wikipedia.org/wiki/Ranking
    .. [2] https://github.com/scipy/scipy/blob/master/scipy/stats/stats.py
    Examples
    --------
    """
    if method not in ('average', 'min', 'max', 'dense', 'ordinal'):
        raise ValueError('unknown method "{0}"'.format(method))

    arr = np.ravel(np.asarray(a))
    algo = 'mergesort' if method == 'ordinal' else 'quicksort'
    sorter = np.argsort(arr, kind=algo)

    inv = np.empty(sorter.size, dtype=np.intp)
    inv[sorter] = np.arange(sorter.size, dtype=np.intp)

    if method == 'ordinal':
        return inv + 1

    arr = arr[sorter]
    obs = np.r_[True, arr[1:] != arr[:-1]]
    dense = obs.cumsum()[inv]

    if method == 'dense':
        return dense

    # cumulative counts of each unique value
    count = np.r_[np.nonzero(obs)[0], len(obs)]

    if method == 'max':
        return count[dense]

    if method == 'min':
        return count[dense - 1] + 1

    # average method
    return .5 * (count[dense] + count[dense - 1] + 1)

def is_eval_allowed(self, cell):
    return not isinstance(cell.compiled_expression, str)
    # isinstance(cell.compiled_expression, str) and re.match('[A-Za-z]+', cell.compiled_expression)

def is_compile_allowed(self, obj):
    return obj.formula is not None

    # This doesn't work in case the value is a
    # string that is the result of a formula like a vlookup, as such that the value
    # will be copied to the compiled expression and used, such that updated will not occur.
    # return not isinstance(obj.value, str)
    #
    # isinstance(self.value, str) and re.match('[A-Za-z]+', self.value):

def create_string_from_cell_value(value):
    # If given value is a String, checks if value contains multiple lines.
    # If value does contain multiple lines, value is split into lines and
    # rebuild to a single String with \\n appended between every line.
    # If value is not a String, just the quotes are added if not already present
    # and the result is returned.
    if not isinstance(value, str):
        raise TypeError('Value is not a String')
    lines = value.splitlines()
    if len(lines) > 1:
        result = '\\n'.join(lines)
    else:
        result = value
    if not result.startswith('"'):
        result = '"' + result
    if not result.endswith('"'):
        result = result + '"'
    return result

def get_zeros(cell):
    if cell.number_format is not None and '.' in cell.number_format:
        # We split the number_format on the dot so that we get a list like so: ['0', '00']
        list_of_zeros = list(cell.number_format.split('.'))
        # Retrieve length of the second index and return it
        length_of_zeros = len(list_of_zeros[1])
        return int(length_of_zeros)
    else: return None


def clean_formula(formula):
    return remove_xlfn(formula)

def remove_xlfn(formula):
    return formula.replace(XLFN, '', 1) if isinstance(formula, str) else formula

def correct_range(excel, rng):
    return [correct_range_single(excel, rg) for rg in rng] if is_collection(rng) else correct_range_single(excel, rng)

def correct_range_single(excel, rng, sheet=None):
    # Ensure the correct sheet is used
    sheet = create_sheet_plus(rng, sheet)
    if sheet is not None:
        excel.set_sheet(sheet)

    # TODONEdNew: retrieve max rows/cols from cache.
    return correct_range_max(rng, excel.max_columns(), excel.max_rows())

def correct_range_max(rng, max_col, max_row):
    rng = rng.replace('$', '')
    sheet, start, end = split_range(rng)
    if not start and not end:
        return rng

    sheet, start, end = split_range(rng)
    sh, start_col, start_row = split_address(start)
    sh, end_col, end_row = split_address(end)

    start_col_nr = col2num(start_col)
    end_col_nr = col2num(end_col)
    if start_col_nr > end_col_nr:
        raise ValueError("Incorrect range specified: [%s], max_col: [%s], max_row: [%s]" % (rng, max_col, max_row))
    mx_col_start = min(start_col_nr, max_col)
    mx_col_end = min(end_col_nr, max_col)

    if start_row is not None:
        start_row_nr = int(start_row)
        mx_row_start = min(start_row_nr, max_row)
    else:
        mx_row_start = 1

    if end_row is not None:
        end_row_nr = int(end_row)
        mx_row_end = min(end_row_nr, max_row)
    else:
        mx_row_end = max_row

    if mx_col_start < start_col_nr or mx_col_end < end_col_nr or start_row is None or mx_row_start < start_row_nr\
            or end_row is None or mx_row_end < end_row_nr:
        # We are requesting more cell and available in the excel, let's correct it
        result = num2col(mx_col_start) if start_col_nr > mx_col_start else str(start_col)
        result += str(mx_row_start) + ':'
        result += num2col(mx_col_end) if end_col_nr > mx_col_end  else str(end_col)
        result += str(mx_row_end)

        #        result =  mx_col_start + str(mx_row_start) + ':' + num2col(mx_col_end) + str(mx_row_end)
        return sheet + '!' + result if sheet else result

    return rng

class CellRange(object):
    def __init__(self, addresses, rng, value, sheet=None):
        self.__addresses = addresses
        self.__address = rng.replace('$', '')
        sh, start, end = split_range(self.__address)

        if not sh and not sheet:
            raise Exception("Must pass in a sheet")

        # make sure the address is always prefixed with the range
        if sh is not None:
            sheet = sh
        else:
            self.__address = sheet + "!" + self.__address

        sh, start_col, start_row = split_address(start)
        sh, end_col, end_row = split_address(end)

        # # don't allow messing with these params

        self.__nrows = int(end_row) - int(start_row)
        self.__ncols = col2num(end_col) - col2num(start_col)

        self.__sheet = sheet
        self.value = value

        # We assume that a new cell range isn't dirty as it's value will be set
        self.__dirty = False
        self.__eval_inprogress = False
        self.__reset_inprogress = False

    def __str__(self):
        return self.__address

    def address(self):
        return self.__address

    """
    Indicates the exact addresses contained in the ranges
    """
    def addresses(self):
        return self.__addresses

    """
    Used to indicate the cell range is dirty, which means re-evaluation is required
    """
    @property
    def dirty(self):
        return self.__dirty

    @dirty.setter
    def dirty(self, value):
        self.__dirty = value

    """
    Used to indicate the cell is being evaluated
    """
    @property
    def eval_inprogress(self):
        return self.__eval_inprogress

    @eval_inprogress.setter
    def eval_inprogress(self, value):
        self.__eval_inprogress = value

    """
    Used to indicate the cell is being reset
    """
    @property
    def reset_inprogress(self):
        return self.__reset_inprogress

    @reset_inprogress.setter
    def reset_inprogress(self, value):
        self.__reset_inprogress = value

    @property
    def nrows(self):
        return self.__nrows

    @property
    def ncols(self):
        return self.__ncols

    # Is required/used when adding a node to the graph
    @property
    def sheet(self):
        return self.__sheet


class Cell(object):
    ctr = 0
    # __dirty = False

    @classmethod
    def next_id(cls):
        cls.ctr += 1
        return cls.ctr


    def __init__(self, address, sheet, formula, value, number_format=None):
        super(Cell, self).__init__()

        # remove $'s
        address = address.replace('$', '')

        sh, c, r = split_address(address)

        # both are empty
        if not sheet and not sh:
            raise Exception("Sheet name may not be empty for cell address %s" % address)
        # both exist but disagree
        elif sh and sheet and sh != sheet:
            raise Exception("Sheet name mismatch for cell address %s: %s vs %s" % (address, sheet, sh))
        elif not sh and sheet:
            sh = sheet
        else:
            pass

        # we assume a cell's location can never change
        # __ Means that it is unchangeable
        self.__sheet = str(sheet)
        self.__sheet = sh
        self.__col = c
        self.__row = int(r) if r else None
        self.__col_idx = col2num(c)

        # Checks if formula is None, if so set formula to None

        if formula is not None:
            self.__formula = formula
            self.value = value
        elif formula is None:
            self.__formula = None
            self.value = value
        elif isinstance(self.value, str):
            self.value = value

        # We assume that a new cell isn't dirty as it's value is set above
        self.__dirty = False
        #  self.__dirty = self.value is None
        self.__eval_inprogress = False
        self.__reset_inprogress = False
        self.number_format = number_format
        self.python_expression = None
        self._compiled_expression = None

        # every cell has a unique id
        self.__id = Cell.next_id()

    @property
    def number_fmt(self):
        return self.number_format

    @number_fmt.setter
    def number_fmt(self, value):
        self.number_format = value

    """
    Used to indicate the cell is dirty, which means re-evaluation is required
    """
    @property
    def dirty(self):
        return self.__dirty

    @dirty.setter
    def dirty(self, value):
        self.__dirty = value

    """
    Used to indicate the cell is being evaluated
    """
    @property
    def eval_inprogress(self):
        return self.__eval_inprogress

    @eval_inprogress.setter
    def eval_inprogress(self, value):
        self.__eval_inprogress = value

    """
    Used to indicate the cell is being reset
    """
    @property
    def reset_inprogress(self):
        return self.__reset_inprogress

    @reset_inprogress.setter
    def reset_inprogress(self, value):
        self.__reset_inprogress = value

    @property
    def sheet(self):
        return self.__sheet

    @property
    def row(self):
        return self.__row

    @property
    def col(self):
        return self.__col

    @property
    def formula(self):
        return self.__formula

    @property
    def id(self):
        return self.__id

    @property
    def compiled_expression(self):
        return self._compiled_expression

    # code objects are not serializable
    def __getstate__(self):
        d = dict(self.__dict__)
        f = '_compiled_expression'
        if f in d:
            del d[f]
        return d

    def __setstate__(self, d):
        self.__dict__.update(d)
        self.compile()

    def clean_name(self):
        return self.address().replace('!', '_').replace(' ', '_')

    def address(self, absolute=True):
        if absolute is True:
            return "%s!%s%s" % (self.__sheet, self.__col, self.__row)
        else:
            return "%s%s" % (self.__col, self.__row)

    def address_parts(self):
        return self.__sheet, self.__col, self.__row, self.__col_idx

    def compile(self):
        if not self.python_expression:
            return

        # if we are a constant string, surround by quotes
        if (isinstance(self.value, str) or isinstance(self.value, int)) \
                and self.formula and self.python_expression.startswith('"'):
            self.python_expression = '"' + self.python_expression + '"'

        try:
            if is_compile_allowed(self, self):
                self._compiled_expression = compile(self.python_expression, '<string>', 'eval')
            else:
                self._compiled_expression = self.value
        except Exception as e:
            raise Exception(
                "Failed to compile cell %s with expression %s: %s" % (self.address(), self.python_expression, e))

    def __str__(self):
        if self.formula is not None:
            return "%s%s" % (self.address(), self.formula)
        else:
            return "%s=%s" % (self.address(), self.value)

    @staticmethod
    def inc_col_address(address, inc):
        sh, col, row = split_address(address)
        return "%s!%s%s" % (sh, num2col(col2num(col) + inc), row)

    @staticmethod
    def inc_row_address(address, inc):
        sh, col, row = split_address(address)
        return "%s!%s%s" % (sh, col, row + inc)

    @staticmethod
    def resolve_cell(excel, address, sheet=None):
        r = excel.get_range(address)
        f = r.Formula if r.Formula.startswith('=') else None
        v = r.value
        nf = r.number_format
        sh, c, r = split_address(address)

        # use the sheet specified in the cell, else the passed sheet
        if sh:
            sheet = sh

        c = Cell(address, sheet, value=v, formula=f, number_format=nf)
        return c

    @staticmethod
    def make_cells_multi(excel: object, ranges: object) -> object:
        result = []
        for sheet, rng in ranges.items():
            # starting points
            cursheet = sheet if sheet else excel.get_active_sheet()
            excel.set_sheet(cursheet)
            rng = correct_range_single(excel, rng,
                                       sheet)  # Ensure the correct sheet is active before correcting the range

            cells, nrows, ncols, rng = Cell.make_cells(excel, rng, cursheet)
            result.extend(list(flatten(cells)))
        return result

    @staticmethod
    def make_cells(excel: object, rng: object, sheet=None, parent_ast=None, cellmap=None) -> object:
        """
        :rtype: object
        """
        cells = []

        def convert_range(rng, sheet=None, parent_ast=None, cellmap=None):
            cells = []

            # use the sheet specified in the range, else the passed sheet
            sh, start, end = split_range(rng)
            if sh is not None:
                sheet = sh

            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Creating cells for range: [%s]", rng)

            ads, numrows, numcols, rgs = resolve_range(rng, sheet=sheet, parent_ast=parent_ast)
            address = create_address(rgs)
            if cellmap is not None and address in cellmap:
                return None, numrows, numcols, address

            if is_collection(rgs):
                rng = rgs
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("Created ranges for range: [%s], ranges: [%s]", rng, rgs)

                fs, vs, nf = [], [], []
                for rn in rgs:
                    r = excel.get_range(rn)
                    fm = r.Formula
                    if is_collection(fm):
                        fs.append(list(flatten(fm)))
                    else:
                        fs.append([fm])

                    nm = r.number_format
                    if is_collection(nf):
                        nf.append(list(flatten(nm)))
                    else:
                        nf.append([nm])

                    va = r.value
                    if is_collection(va):
                        vs.append(list(flatten(va)))
                    else:
                        vs.append([va])

                if len(fs) != len(ads):
                    fs = list(zip(*fs))
                    vs = list(zip(*vs))
                    nf = list(zip(*nf))
            else:
                r = excel.get_range(rng)
                fs = r.Formula
                vs = r.value
                nf = r.number_format  # nf: NumberFormat
                # It must be a 2-dim array (row and column)
                if numrows == 1:
                    ads = [ads]
                elif numcols == 1:
                    ads = [[x] for x in ads]

            # In case the value is a single integer, we must make it a string else the zip will throw an error indicating
            # it's not iteratable.
            # In case vs is None we have to wrap it in an array as well, else the zip function below will fail as None
            # isn't iteratable
            if vs is None or isinstance(vs, (int, float, str)):
                vs = [[str(vs)]]

            # We put the formula in a list, as a single string might result in problems.
            # Encountered problem was: ads=[['AV5']], fs='=AV178' that resulted below in the zip in fs='='
            # Apparently unzipping (zip(*x)) will use the "=" as some kind of separator.
            if fs is None or isinstance(fs, str):
                fs = [[fs]]

            if nf is None or isinstance(nf, str):
                nf = [[nf]]

            try:
                # We can't merge ads, fs and vs as vs can be less that can result in cells being ignored
                # Example: last column contains only formulas such that vs returns rows with 1 column less then a fs row
                for ad_row_nr in range(len(ads)):
                    row = []
                    ad_row = ads[ad_row_nr]
                    for ad_cell_nr in range(len(ad_row)):
                        a = ad_row[ad_cell_nr]
                        if is_collection(a):
                            raise Exception("Collection not allowed: [%s], range: [%s], ranges: [%s], formulas: [%s], "
                                            "values: [%s]" % (a, rng, rgs, fs, vs))

                        formula = clean_formula(fs[ad_row_nr][ad_cell_nr])
                        formula = formula if formula and formula.startswith('=') else None

                        value = vs[ad_row_nr]
                        nr_format = nf[ad_row_nr]

                        # A value can be empty, example: the last column contains only formulas
                        value = value[ad_cell_nr] if ad_cell_nr < len(value) else None
                        nr_format = nr_format[ad_cell_nr] if ad_cell_nr < len(nr_format) else None
                        cl = Cell(a, sheet, value=value, formula=formula, number_format=nr_format)

                        # We expect a formula that is at least 2 character long, the first character should be the "="
                        if formula is not None and len(formula) < 2:
                            raise Exception(
                                "Wrong formula encountered: [%s], a: [%s], v: [%s], ads: [%s], fs: [%s], "
                                "vs: [%s], range: [%s], cell: [%s]" % (formula, a, value, ads, fs, vs, rng, cl))
                        row.append(cl)

                        if logger.isEnabledFor(LOG_FINE):
                            logger.log(LOG_FINE, "Created cell for range: [%s!%s], cell: [%s]", sheet, rng, cl)

                    if len(row) > 0:
                        cells.append(row)

                # return as vector
                if len(cells) > 0 and (numrows == 1 or numcols == 1):
                    if numrows == 1:
                        cells = cells[0]
                    else:
                        cls = []
                        for x in cells:
                            # We ignore empty cells, like cells that result in an excel #REF error.
                            if x is not None:
                                cls.append(x[0] if isinstance(x, list) else x)

                                # Removed: as it creates an error for empty/removed (#REF excel error) cells
                                # cells = [x[0] for x in cells]
                        cells = cls

                return cells, numrows, numcols, rng
            except Exception as ex:
                raise Exception(
                    "Error creating row cells for range [%s], sheet: [%s], ads: [%s], num_rows: [%s], num_cols: [%s], formula: [%s],"
                    " values: [%s], cells: [%s], number_format: [%s]" % (rng, sheet, ads, numrows, numcols, fs, vs, cells, nr_format)) from ex

        def create_range_exception(rng, cell, sheet):
            return create_exception(rng, cell, sheet, "Error converting")

        def create_resolve_exception(rng, cell, sheet):
            return create_exception(rng, cell, sheet, "Error resolving")

        def create_exception(rng, cell, sheet, msg):
            return Exception(msg + ", range: [%s], cell: [%s], sheet: [%s], excel: [%s]"
                             % (rng, cell, sheet, excel))

        if isinstance(rng, list):  # if a list of cells
            cells = []
            for cell in rng:
                if is_range(cell):
                    try:
                        cs_in_range, nr, nc, rng = convert_range(cell, sheet, cellmap)
                        cells.append(cs_in_range)
                    except Exception as ex:
                        raise create_range_exception(rng, cell, sheet) from ex
                else:
                    try:
                        c = Cell.resolve_cell(excel, cell, sheet=sheet)
                        cells.append(c)
                    except Exception as ex:
                        raise create_resolve_exception(rng, cell, sheet) from ex

            cells = list(flatten(cells))
            return cells, -1, -1, None  # numrows, numcols and range are irrelevant here, so we return nr=nc=-1
        else:
            if is_range(rng):
                try:
                    # We return the range as it will be specific in case the shorthand range notitation was used,
                    # shorthand: (A:B), and specific: (A88:B99)
                    cells, numrows, numcols, rng = convert_range(rng, sheet, parent_ast, cellmap)
                except Exception as ex:
                    raise create_range_exception(rng, None, sheet) from ex
            else:
                try:
                    c = Cell.resolve_cell(excel, rng, sheet=sheet)
                    cells.append(c)
                    numrows = 1
                    numcols = 1
                except Exception as ex:
                    raise create_resolve_exception(rng, None, sheet) from ex

        return cells, numrows, numcols, rng

def is_range(address):
    return address.find(':') > 0

def split_range(rng):
    try:
        sh, start, end = None, None, None
        if rng.find('!') > 0:
            sh, r = rng.split("!")
            if rng.find(':') > 0:
                start, end = r.split(':')
        else:
            sh = None
            if rng.find(':') > 0:
                start, end = rng.split(':')

        return sh, start, end
    except Exception as ex:
        raise Exception("Error spliting range: [%s]" % (rng)) from ex

def create_sheet(rng):
    if rng.find('!') > 0:
        sh, r = rng.split("!")
        return sh
    return None

"""
It includes the '!' symbol at the end.
In case the value of sheet is specified, it must match the sheet value in the specified range.
We use the specified sheet, in case it's present and not present in the specified range.
"""
def create_sheet_plus(rng, sheet=None, add_close=False):
    sh, start, end = split_range(rng)
    if sh and sheet is not None:
        if sh != sheet:
            sheet = sh
            # sh = sheet
            # raise Exception("Mismatched sheets %s and %s" % (sh, sheet))
    elif sh is not None and not sheet:
        sheet = sh

    if sheet and add_close:
        sheet += '!'
    return sheet

def split_address(address):
    sheet = None
    if address.find('!') > 0:
        sheet, address = address.split('!')

    # ignore case
    address = address.upper()

    # regular <col><row> format    
    if re.match('^[A-Z $]+[\d $]+$', address):
        col, row = [_f for _f in re.split('([A-Z $]+)', address) if _f]
    # shorthand range notition (A:H) instead of (A1:H100) (al rows in A till all rows in H)
    elif re.match('^[A-Z $]+$', address):
        col = address
        row = None
    # R<row>C<col> format
    elif re.match('^R\d+C\d+$', address):
        row, col = address.split('C')
        row = row[1:]
    # R[<row>]C[<col>] format
    elif re.match('^R\[\d+]C\[\d+]$', address):
        row, col = address.split('C')
        row = row[2:-1]
        col = col[2:-1]
    else:
        raise Exception('Invalid address format ' + address)

    return sheet, col, row

def create_range(sheet, start_col, start_row, end_col, end_row):
    if not start_col or not end_col:
        raise Exception("Columns can't be None, start_col: [%s], start_row: [%s], end_col: [%s], "
                        "end_row: [%s]" % (start_col, start_row, end_col, end_row))

    result = start_col
    if start_row is not None:
        result += str(start_row)
    result += ':' + end_col
    if end_row is not None:
        result += str(end_row)
    if sheet is not None:
        if not sheet.find('!') > 0:
            sheet += '!'
        result = sheet + result
    return result

def resolve_function(formula):
    if formula and formula.startswith('='):
        fn, r = formula.split("(")
        return fn[1:]

def find_key_by_nested_key_value(dicts, deps, ky, value):
    for dep in deps:
        node = dicts[dep]
        if ky in node and value == node[ky]:
            return dep
    return None

def resolve_range_vlookup(rng, sheet, rng_node, formula_node, parent_ast, flatten=False):
    if not is_range(rng):
        return resolve_range_simple(rng, sheet)

    sheet = safe_str(create_sheet_plus(rng, sheet))
    sh, start, end = split_range(rng)
    sh, start_col, start_row = split_address(start)
    sh, end_col, end_row = split_address(end)
    search = find_key_by_nested_key_value(parent_ast.node, parent_ast.pred[formula_node], 'pos', 2)
    try:
        col = num2col(col2num(start_col) + int(search.tvalue) - 1)
    except Exception as ex:
        if not search:
            raise Exception("Search value not-found/incorrect, search: [%s], range: [%s], formula: "
                            "[%s]" % (search, rng, formula_node))
        # do it the normal way
        if logger.isEnabledFor(logging.INFO):
            logger.info("Unable to restrict the vlookup range: [%s], formula node: [%s]", rng, formula_node)

        return resolve_range_simple(rng, sheet)

    # Adding the search col
    first = create_range(sheet, start_col, start_row, start_col, end_row)
    addr1, nrows, ncols, rg = resolve_range(first, sheet=sheet, flatten=flatten)

    # Adding the result col
    second = create_range(sheet, col, start_row, col, end_row)
    addr2, nrows, ncols, rg = resolve_range(second, sheet=sheet, flatten=flatten)

    if logger.isEnabledFor(logging.DEBUG):
        logger.debug("Vlookup range: [%s] optimized to: [%s]", rng, [first, second])

    addr1 = list(zip(addr1, addr2))
    if flatten:
        addr1 = flatten(addr1)
        return addr1, 1, len(addr1)

    return addr1, nrows, 2, [first, second]

def resolve_range_hlookup(rng, sheet, rng_node, formula_node, parent_ast, flatten):
    if not is_range(rng):
        return resolve_range_simple(rng, sheet)

    sheet = safe_str(create_sheet_plus(rng, sheet))
    sh, start, end = split_range(rng)
    sh, start_col, start_row = split_address(start)
    sh, end_col, end_row = split_address(end)
    search = find_key_by_nested_key_value(parent_ast.node, parent_ast.pred[formula_node], 'pos', 2)
    try:
        row = int(start_row) + int(search.tvalue) - 1
    except Exception as ex:
        if not search:
            raise Exception("Search value not-found/incorrect, search: [%s], range: [%s], formula: "
                            "[%s]" % (search, rng, formula_node))
        # do it the normal way
        if logger.isEnabledFor(logging.INFO):
            logger.info("Unable to restrict the hlookup range: [%s], formula node: [%s]", rng, formula_node)
        return resolve_range_simple(rng, sheet)

    # Adding the search row
    first = create_range(sheet, start_col, start_row, end_col, start_row)
    addr1, nrows, ncols, rg = resolve_range(first, sheet=sheet, flatten=flatten)

    # Adding the second/result row
    second = create_range(sheet, start_col, row, end_col, row)
    addr2, nrows, ncols, rg = resolve_range(second, sheet=sheet, flatten=flatten)

    if logger.isEnabledFor(logging.DEBUG):
        logger.debug("Hlookup range: [%s] optimized to: [%s]", rng, [first, second])

    if flatten:
        addr1 = flatten([addr1, addr2])
        return addr1, 1, len(addr1)
    return [addr1, addr2], 2, ncols, [first, second]

def resolve_range(rng, sheet=None, flatten=False, parent_ast=None):
    # single cell, no range
    if not is_range(rng):
        return [sheet + rng], 1, 1

    if parent_ast:
        # if rng.find('!') > 0:
        #     s, rg = rng.split("!")
        # We assume that the specified range is the same as used as key in the ast graph. In case the sheet is missing
        # and is present in the ast graph, we can't find it.
        rng_node = find_range_node(parent_ast, rng)
        if rng_node:
            formula_node = find_formula_node(rng_node, parent_ast)
            if formula_node:
                if formula_node.tvalue.lower() == 'vlookup':
                    return resolve_range_vlookup(rng, sheet, rng_node, formula_node, parent_ast, flatten)
                elif formula_node.tvalue.lower() == 'hlookup':
                    return resolve_range_hlookup(rng, sheet, rng_node, formula_node, parent_ast, flatten)
    return resolve_range_simple(rng, sheet, flatten)

def find_range_node(parent_ast, rg):
    from .excel_node import RangeNode
    for node in parent_ast.nodes():
        if isinstance(node, RangeNode):
            if node.tvalue.replace('$', '') == rg:
                return node

                # nodes = [x for x in parent_ast.nodes() if isinstance(x, RangeNode) and x.tvalue == rg]
                # return nodes[0] if len(nodes) else None

def find_formula_node(rng_node, parent_ast):
    formula_node = parent_ast.succ[rng_node]
    if len(formula_node) > 1:
        raise Exception(
            "Only one formula expected: [%s], range: [%s], parent_ast: [%s]" % (formula_node, rng_node, parent_ast))

    return next(iter(formula_node))  # retrieve the first value in the dict

def resolve_range_simple(rng, sheet, flatten=False):
    sh, start, end = split_range(rng)
    sh, start_col, start_row = split_address(start)
    sh, end_col, end_row = split_address(end)

    if not start_row or not end_row or not start_col or not end_col:
        # As this point we assume the ranges are corrected such that the range contains a full start/end row with the
        # row number. Example: range "E:G" should be corrected to "E1:G234" using the max row/columns of a sheet.
        raise Exception(
            "Row/Col start/end missing, range: [%s], start row: [%s], end row: [%s], start col: [%s], end col: [%s]"
            % (rng, start_row, end_row, start_col, end_col))

    # Let's return some "marker" values we use to recognize the shorthand notition when required.
    # if start_row is None and end_row is None:
    #     return 'all', -1, -1

    start_row = int(start_row)
    end_row = int(end_row)

    if start_col == end_col:  # single column
        return resolve_range_col(rng, start_row, end_row, start_col, sheet)
    elif start_row == end_row:  # single row
        return resolve_range_row(rng, start_row, end_row, start_col, end_col, sheet)
    else:  # rectangular range
        return resolve_range_rectangle(rng, start_row, end_row, start_col, end_col, sheet, flatten)

def resolve_range_rectangle(rng, start_row, end_row, start_col, end_col, sheet, flatten):
    sheet = create_sheet_plus(rng, sheet, True)
    start_col_nr = col2num(start_col)
    end_col_nr = col2num(end_col)
    cells = []
    for r in range(start_row, end_row + 1):
        row = []
        for c in range(start_col_nr, end_col_nr + 1):
            row.append(safe_str(sheet) + num2col(c) + str(r))
        cells.append(row)

    if flatten:
        # flatten into one list
        l = flatten(cells)
        return l, 1, len(l), rng
    else:
        return cells, len(cells), len(cells[0]), rng

def resolve_range_row(rng, start_row, end_row, start_col, end_col, sheet):
    sheet = create_sheet_plus(rng, sheet, True)
    start_col_nr = col2num(start_col)
    end_col_nr = col2num(end_col)
    ncols = end_col_nr - start_col_nr + 1
    data = ["%s%s%s" % (s, num2col(c), r) for (s, c, r) in
            zip([safe_str(sheet)] * ncols, list(range(start_col_nr, end_col_nr + 1)), [start_row] * ncols)]
    return data, 1, len(data), rng

def resolve_range_col(rng, start_row, end_row, start_col, sheet):
    sheet = create_sheet_plus(rng, sheet, True)
    nrows = end_row - start_row + 1
    data = ["%s%s%s" % (s, c, r) for (s, c, r) in
            zip([safe_str(sheet)] * nrows, [start_col] * nrows, list(range(start_row, end_row + 1)))]
    return data, len(data), 1, rng

def col2num(col):
    if not col:
        raise Exception("Column may not be empty")

    tot = 0
    for i, c in enumerate([c for c in col[::-1] if c != "$"]):
        if c == '$':
            continue
        tot += (ord(c) - 64) * 26 ** i
    return tot

# convert back
def num2col(num):
    if num < 1:
        raise Exception("Number must be larger than 0: %s" % num)

    s = ''
    q = num
    while q > 0:
        (q, r) = divmod(q, 26)
        if r == 0:
            q = q - 1
            r = 26
        s = string.ascii_uppercase[int(r - 1)] + s
    return s

def address2index(a):
    sh, c, r = split_address(a)
    return col2num(c), int(r)

def index2addres(c, r, sheet=None):
    return "%s%s%s" % (sheet + "!" if sheet else "", num2col(c), r)

def get_linest_degree(excel, cl):
    # TODO: assumes a row or column of linest formulas & that all coefficients are needed

    sh, c, r, ci = cl.address_parts()
    # figure out where we are in the row

    # to the left
    i = ci - 1
    while i > 0:
        f = excel.get_formula_from_range(index2addres(i, r))
        if f is None or f != cl.formula:
            break
        else:
            i = i - 1

    # to the right
    j = ci + 1
    while True:
        f = excel.get_formula_from_range(index2addres(j, r))
        if f is None or f != cl.formula:
            break
        else:
            j = j + 1

    # assume the degree is the number of linest's
    degree = (j - i - 1) - 1  # last -1 is because an n degree polynomial has n+1 coefs

    # which coef are we (left most coef is the coef for the highest power)
    coef = ci - i

    # no linests left or right, try looking up/down
    if degree == 0:
        # up
        i = r - 1
        while i > 0:
            f = excel.get_formula_from_range("%s%s" % (c, i))
            if f is None or f != cl.formula:
                break
            else:
                i = i - 1

        # down
        j = r + 1
        while True:
            f = excel.get_formula_from_range("%s%s" % (c, j))
            if f is None or f != cl.formula:
                break
            else:
                j = j + 1

        degree = (j - i - 1) - 1
        coef = r - i

    # if degree is zero -> only one linest formula -> linear regression -> degree should be one
    return max(degree, 1), coef

def flatten(l):
    if is_collection(l):
        for el in l:
            if isinstance(el, collections.Iterable) and not isinstance(el, str):
                for sub in flatten(el):
                    yield sub
            else:
                yield el
    return l

def uniqueify(seq):
    seen = set()
    seen_add = seen.add
    return [x for x in seq if x not in seen and not seen_add(x)]

def is_valid(value, criteria):
    check = criteria_parser(criteria)
    return check(value)

def is_collection(args):
    return isinstance(args, (list, tuple))

def is_nr(value):
    return value is not None and isinstance(value, (int, float))

def is_number(s):  # http://stackoverflow.com/questions/354038/how-do-i-check-if-a-string-is-a-number-float-in-python
    try:
        float(s)
        return True
    except ValueError:
        return False

def is_leap_year(year):
    if not is_number(year):
        raise TypeError("%s must be a number" % str(year))
    if year <= 0:
        raise TypeError("%s must be strictly positive" % str(year))

    # Watch out, 1900 is a leap according to Excel => https://support.microsoft.com/en-us/kb/214326
    return (year % 4 == 0 and year % 100 != 0 or year % 400 == 0) or year == 1900

def get_max_days_in_month(month, year):
    if not is_number(year) or not is_number(month):
        raise TypeError("All inputs must be a number")
    if year <= 0 or month <= 0:
        raise TypeError("All inputs must be strictly positive")

    if month in (4, 6, 9, 11):
        return 30
    elif month == 2:
        if is_leap_year(year):
            return 29
        else:
            return 28
    else:
        return 31

def normalize_year(y, m, d):
    if m <= 0:
        y -= int(abs(m) / 12 + 1)
        m = 12 - (abs(m) % 12)
        normalize_year(y, m, d)
    elif m > 12:
        y += int(m / 12)
        m = m % 12

    if d <= 0:
        d += get_max_days_in_month(m, y)
        m -= 1
        y, m, d = normalize_year(y, m, d)

    else:
        if m in (4, 6, 9, 11) and d > 30:
            m += 1
            d -= 30
            y, m, d = normalize_year(y, m, d)
        elif m == 2:
            if (is_leap_year(y)) and d > 29:
                m += 1
                d -= 29
                y, m, d = normalize_year(y, m, d)
            elif (not is_leap_year(y)) and d > 28:
                m += 1
                d -= 28
                y, m, d = normalize_year(y, m, d)
        elif d > 31:
            m += 1
            d -= 31
            y, m, d = normalize_year(y, m, d)

    return y, m, d

def date_from_int(nb):
    if not is_number(nb):
        raise TypeError("%s is not a number" % str(nb))

    # origin of the Excel date system
    current_year = 1900
    current_month = 0
    current_day = 0

    while (nb > 0):
        if not is_leap_year(current_year) and nb > 365:
            current_year += 1
            nb -= 365
        elif is_leap_year(current_year) and nb > 366:
            current_year += 1
            nb -= 366
        else:
            current_month += 1
            max_days = get_max_days_in_month(current_month, current_year)

            if nb > max_days:
                nb -= max_days
            else:
                current_day = nb
                nb = 0

    return current_year, current_month, current_day

def criteria_parser(criteria):
    if is_number(criteria):
        def check(x):
            return x == criteria  # and type(x) == type(criteria)
    elif type(criteria) == str:
        search = re.search('(\W*)(.*)', criteria.lower()).group
        operator = search(1)
        value = search(2)
        value = float(value) if is_number(value) else str(value)

        if operator == '<':
            def check(x):
                if not is_number(x):
                    return False  # Added such that it skips non-numbers in for example the sumif function
                    # raise TypeError(
                    #     'excellib.countif() doesnt\'t work for checking non number items against non equality')
                return x < value
        elif operator == '>':
            def check(x):
                if not is_number(x):
                    return False
                    # raise TypeError(
                    #     'excellib.countif() doesnt\'t work for checking non number items against non equality')
                return x > value
        elif operator == '>=':
            def check(x):
                if not is_number(x):
                    return False
                    # raise TypeError(
                    #     'excellib.countif() doesnt\'t work for checking non number items against non equality')
                return x >= value
        elif operator == '<=':
            def check(x):
                if not is_number(x):
                    return False
                    # raise TypeError(
                    #     'excellib.countif() doesnt\'t work for checking non number items against non equality')
                return x <= value
        elif operator == '<>':
            def check(x):
                if not is_number(x):
                    return False
                    # raise TypeError(
                    #     'excellib.countif() doesnt\'t work for checking non number items against non equality')
                return x != value
        else:
            def check(x):
                return x == criteria
    else:
        raise Exception('Could\'t parse criteria %s' % criteria)

    return check

def find_corresponding_index(rng, criteria):
    # parse criteria
    check = criteria_parser(criteria)
    valid = []
    for index, item in enumerate(rng):
        if check(item) is not None:
            valid.append(index)

    return valid

def correct_numbers(val1, val2):
    try:
        if val1 == '':
            val1 = None
        elif val2 == '':
            val2 = None
        elif isinstance(val1, str):
            val1 = float(val1)
        elif isinstance(val2, str):
            val2 = float(val2)

        return val1, val2

    except Exception as ex:
        if val1 or val2:
            return None, None
        raise Exception("Incorrect values for val1=[%s], val2=[%s]" % (val1, val2)) from ex


if __name__ == '__main__':
    pass



"""
Python equivalents of various excel functions
"""

import math
import re
import numpy as np
from datetime import datetime
from math import log
from decimal import Decimal, ROUND_HALF_UP
from converter.excel_util import flatten, is_number, is_collection, date_from_int, normalize_year, is_leap_year, \
    find_corresponding_index, is_nr, ERR_NA, ERR_DIV, ERR_VALUE, ERROR_VALUES, rankdata, split_range, split_address, \
    col2num, num2col, evaluate_spreadsheet, to_values, is_valid, iterate_cells, get_range, safe_str as util_safe_str, \
    correct_numbers

# #####################################################################################
# A dictionary that maps excel function names onto python equivalents. You should
# only add an entry to this map if the python name is different to the excel name
# (which it may need to be to prevent conflicts with existing python functions
# with that name, e.g., max).

# So if excel defines a function foobar(), all you have to do is add a function
# called foobar to this module.  You only need to add it to the function map,
# if you want to use a different name in the python code. 

# Note: some functions (if, pi, atan2, and, or, array, ...) are already taken care of
# in the FunctionNode code, so adding them here will have no effect.
# Only add functions that have a different name in this file then the one in excel.
FUNCTION_MAP = {
    "ln": "xlog",
    "min": "xmin",
    "max": "xmax",
    "sum": "xsum",
    "gammaln": "lgamma",
    "round": "xround",
    "rank.eq": "rank_eq",
}

######################################################################################
# List of excel equivalent functions
# TODO: needs unit testing
######################################################################################
def greater_then(val1, val2, equals):
    # Excel seems to treat "None" value's like if they are '0'
    if val1 is None or isinstance(val1, str) and len(val1) == 0:
        val1 = 0
    if val2 is None or isinstance(val2, str) and len(val2) == 0:
        val2 = 0

    if val1 in ERROR_VALUES:
        return val1
    if val2 in ERROR_VALUES:
        return val2

    if isinstance(val1, str) or isinstance(val2, str):
        val1 = str(val1)
        val2 = str(val2)

    return val1 >= val2 if equals else val1 > val2

def less_then(val1, val2, equals):
    # Excel seems to treat "None" value's like if they are '0'
    if val1 is None or isinstance(val1, str) and len(val1) == 0:
        val1 = 0
    if val2 is None or isinstance(val2, str) and len(val2) == 0:
        val2 = 0

    if val1 in ERROR_VALUES:
        return val1
    if val2 in ERROR_VALUES:
        return val2

    if isinstance(val1, str) or isinstance(val2, str):
        val1 = str(val1)
        val2 = str(val2)

    return val1 <= val2 if equals else val1 < val2

def minus(val1, val2):
    if val1 in ERROR_VALUES:
        return val1
    if val2 in ERROR_VALUES:
        return val2
    val1, val2 = correct_numbers(val1, val2)

    return val1 - val2 if val1 is not None and val2 is not None else val1 if val1 is not None else val2

def plus(val1, val2):
    if val1 in ERROR_VALUES:
        return val1
    if val2 in ERROR_VALUES:
        return val2
    val1, val2 = correct_numbers(val1, val2)

    return val1 + val2 if val1 is not None and val2 is not None else val1 if val1 is not None else val2

def multiply(val1, val2):
    if val1 in ERROR_VALUES:
        return val1
    if val2 in ERROR_VALUES:
        return val2
    val1, val2 = correct_numbers(val1, val2)
    if val1 is None or val2 is None:
        return ERR_VALUE
    return val1 * val2 if val1 is not None and val2 is not None else 0

def divide(val1, val2):
    if val1 in ERROR_VALUES:
        return val1
    if val2 in ERROR_VALUES:
        return val2
    val1, val2 = correct_numbers(val1, val2)
    return val1 / val2 if val1 is not None and val2 is not None else 0 if val1 is None else ERR_DIV

def safe_str(value):
    return util_safe_str(value)

def sin(val1, args):
    if not is_nr(val1) or val1 == 0:
        raise Exception("The value must be a Number [%s]" % val1)
    else:
        return math.sin(val1)

def row(address):
    col, row = [_f for _f in re.split('([A-Z $]+)', address) if _f]
    return int(row)

def sumproduct(spreadsheet, main_address, *args):
    result = 0
    if len(args) > 1:
        # Not sure exactly how it works with more then 1 array in args (2 or more in total)
        raise Exception("Currently we don't support more then 2 arrays in total. Address: [%s]" % main_address)

    rng = get_range(spreadsheet, main_address)
    sheet, start, end = split_range(rng)
    sh, start_col, start_row = split_address(start)
    start_col_nr = col2num(start_col)

    def process(col_nr, row_nr, value):
        if value is not None:
            nonlocal result
            if len(args):
                for ar in args:
                    sheet, start, end = split_range(ar)
                    sh, start_cl, start_rw = split_address(start)
                    try:
                        val = evaluate_spreadsheet(spreadsheet, sheet,
                                                   num2col(col2num(start_cl) + col_nr - start_col_nr),
                                                   int(start_rw) + row_nr - int(start_row))
                        if val is not None:
                            result += value * val
                    except:
                        result = ERR_VALUE
                        return False
            else:
                if value is not None:
                    result += value
        return True

    iterate_cells(spreadsheet, rng, process)
    return result

def substitute(spreadsheet, source, old, new, nth=None):
    if nth == 0:
        return ERR_VALUE

    if nth is not None:  # not allowed to b
        find = source.find(old)
        # if find is not p1 we have found at least one match for the substring
        i = find != -1
        # loop util we find the nth or we find no match
        while find != -1 and i != nth:
            # find + 1 means we start at the last match start index + 1
            find = source.find(old, find + 1)
            i += 1
        # if i is equal to nth we found nth matches so replace
        if i == nth:
            return source[:find] + new + source[find + len(old):]
        return source
    else:
        return source.replace(old, new)

def xlog(a):
    if isinstance(a, (list, tuple, np.ndarray)):
        return [log(x) for x in flatten(a)]
    else:
        return log(a)

def xmax(spreadsheet, *args):
    result = None
    if is_collection(args) is not None:
        items = [x for x in args if is_nr(x)]
        if len(items):  # it concerns a collection of numbers
            return max(items)

        else:  # it concerns a nested range, collection,...
            for arg in args:
                sub = xmax_single(spreadsheet, arg)
                if is_nr(sub):
                    result = min(result, sub) if is_nr(result) else sub
    else:
        result = xmax_single(spreadsheet, args)
    return result if result is not None else ERR_NA

def xmax_single(spreadsheet, arg):
    result = None
    if is_nr(arg):
        result = max(result, arg) if is_nr(result) else arg
    elif isinstance(arg, str):
        def process(col_nr, row_nr, value):
            nonlocal result
            if is_nr(value) and (result is None or value > result):
                result = value
            return True

        rng = get_range(spreadsheet, arg)
        iterate_cells(spreadsheet, rng, process)
    else:
        return ERR_VALUE

    return result if result is not None else ERR_NA

def xmin(spreadsheet, *args):
    result = None
    if is_collection(args):
        items = [x for x in args if is_nr(x)]
        if len(items):  # it concerns a collection of numbers
            return min(items)

        else:  # it concerns a nested range, collection,...
            for arg in args:
                sub = xmin_single(spreadsheet, arg)
                if is_nr(sub):
                    result = min(result, sub) if is_nr(result) else sub
    else:
        result = xmin_single(spreadsheet, args)
    return result if result is not None else ERR_NA

def xmin_single(spreadsheet, arg):
    result = None
    if is_nr(arg):
        result = min(result, arg) if is_nr(result) else arg
    elif isinstance(arg, str):
        def process(col_nr, row_nr, value):
            nonlocal result
            if is_nr(value) and (result is None or value < result):
                result = value
            return True

        rng = get_range(spreadsheet, arg)
        iterate_cells(spreadsheet, rng, process)
    else:
        return ERR_VALUE

    return result if result is not None else ERR_NA

def xsum(spreadsheet, *args):
    result = 0
    if args in ERROR_VALUES:
        return args

    def process(col_nr, row_nr, value):
        nonlocal result
        if value is not None and is_nr(value):
            result += value
        return True

    if is_nr(args):
        return args

    # Ensure to use the exact range
    rng = None
    if is_collection(args):
        rng = []
        for ad in args:
            if is_nr(ad):
                result += ad
            else:
                rng.append(get_range(spreadsheet, ad))
    else:
        rng = args
    if is_collection(args):
        iterate_cells(spreadsheet, rng, process)
    # result = to_values(spreadsheet, rng, True, is_nr);
    return result

"""
    Ref: https://support.office.com/en-us/article/SUMIF-function-169b8c99-c05c-4483-a712-1697a653039b
"""
def sumif(spreadsheet, cell_address, criteria, sum_range=None):
    rng = get_range(spreadsheet, cell_address)

    def is_valid_cell(value):
        return is_valid(value, criteria)

    return sum(to_values(spreadsheet, rng, True, is_valid_cell))

def average(spreadsheet, cell_address):
    result = 0
    counter = 0

    def process(col_nr, row_nr, value):
        nonlocal result, counter
        if value is not None and not is_nr(value):
            return ERR_NA
        else:
            result += value
            counter += 1
        return True

    # Ensure to use the exact range
    rng = get_range(spreadsheet, cell_address)
    iterate_cells(spreadsheet, rng, process)
    return result / counter

def right(spreadsheet, text, n):
    if isinstance(text, str) or isinstance(text, str):
        return text[-n:]
    else:
        # TODO: get rid of the decimal
        return str(int(text))[-n:]

"""
Ref: https://support.office.com/en-us/article/INDEX-function-a5dcf0dd-996d-40a4-a822-b56b061328bd
"""
def index(spreadsheet, cell_address, row_nr=None, col_nr=None):
    try:
        return index_inter(spreadsheet, cell_address, row_nr, col_nr)
    except Exception as ex:
        raise Exception(
            "Error in index function, cell_address: [%s], row_nr: [%s], col_nr: [%s]" % (cell_address, row_nr, col_nr))

def index_inter(spreadsheet, cell_address, row_nr=None, col_nr=None):
    if row_nr is None and col_nr is None or row_nr is not None and row_nr == 0 or col_nr is not None and col_nr == 0:
        return ERR_VALUE
    elif row_nr in ERROR_VALUES:
        return row_nr

    rng = get_range(spreadsheet, cell_address)
    sheet, start, end = split_range(rng)
    sh, start_col, start_row = split_address(start)
    if row_nr is not None and col_nr is not None:
        col = num2col(col2num(start_col) + col_nr - 1)
        row = int(start_row) + int(row_nr - 1)
    elif row_nr is not None:
        sh, end_col, end_row = split_address(end)
        if start_col == end_col:  # it concerns a vertical array
            col = start_col
            row = int(start_row) + int(row_nr) - 1
        elif start_row == end_row:  # it concerns a horizontal array
            col = num2col(col2num(start_col) + row_nr - 1)
            row = int(start_row)
        else:
            raise Exception(
                "Multi dimension array not allowed with only row or column, cell_address: [%s], col_nr: [%s], row_nr: [%s]" % (
                    cell_address, col_nr, row_nr))
    else:
        return ERR_VALUE
        # raise Exception("Both row_nr and col_nr are missing, cell_address: [%s]" % (cell_address))
    return evaluate_spreadsheet(spreadsheet, sheet, col, int(row))

"""
Ref: https://support.office.com/en-us/article/RANK-EQ-function-284858ce-8ef6-450e-b662-26245be04a40
"""
def rank_eq(spreadsheet, item, cell_address, mode):
    rng = get_range(spreadsheet, cell_address)
    sheet, start, end = split_range(rng)
    sh, start_col, start_row = split_address(start)
    sh, end_col, end_row = split_address(end)
    start_col_nr, start_row_nr = col2num(start_col), int(start_row)
    end_row_nr = int(end_row)
    if not item or item == ERR_NA:
        return ERR_NA
    if start_col != end_col:
        raise Exception("The start and end column are expected to be the same")

    items = []
    for row_nr in range(start_row_nr, end_row_nr + 1):
        value = evaluate_spreadsheet(spreadsheet, sheet, num2col(start_col_nr), row_nr)
        if value is not None:
            items.append(value)
    if mode == 1:
        return rankdata(items, 'min')[items.index(item)]
    else:
        raise Exception("Mode isn't supported yet, mode: [%s], item: [%s]" % (mode, item))

def lookup(value, lookup_range, result_range):
    if not is_nr(value):
        raise Exception("Non numeric lookups (%s) not supported" % value)

    # TODO: note, may return the last equal value

    # index of the last numeric value
    lastnum = -1
    for i, v in enumerate(lookup_range):
        if isinstance(v, (int, float)):
            if v > value:
                break
            else:
                lastnum = i

    if lastnum < 0:
        raise Exception("No numeric data found in the lookup range")
    else:
        if i == 0:
            raise Exception("All values in the lookup range are bigger than %s" % value)
        else:
            if i >= len(lookup_range) - 1:
                # return the biggest number smaller than value
                return result_range[lastnum]
            else:
                return result_range[i - 1]

"""
The VLOOKUP function performs a vertical lookup by searching for a value in the first column of a table and 
returning the value in the same row in the index_number position.
Ref: https://www.techonthenet.com/excel/formulas/vlookup.php
"""
def vlookup(spreadsheet, item, cell_address, col_index_nr, approximate_match=False):
    if not item or item == ERR_NA:
        return ERR_NA

    if approximate_match and not isinstance(item, (float, int)):
        raise Exception("Item must be a number: [%s], row_index_nr: [%s]" % (item, col_index_nr))

    # We assume the range has been correctly Removed.
    # we can't find it in the cellmap as we changed the cell address to overcome conflict with other cells using
    # the same range
    # rng = get_range(spreadsheet, cell_address)
    rng = cell_address
    item = str(item).lower()
    sheet, start, end = split_range(rng)
    sh, start_col, start_row = split_address(start)
    sh, end_col, end_row = split_address(end)
    search_index = num2col(col2num(start_col) + col_index_nr - 1)
    previous = None
    ln = int(end_row) - int(start_row) + 1
    for row_nr in range(int(start_row), int(end_row) + 1):
        value = evaluate_spreadsheet(spreadsheet, sheet, start_col, row_nr)
        if value is None:  # we skip empty cells/columns
            continue
        value = str(value).lower()
        if approximate_match and previous is not None and int(previous) > int(value):
            raise Exception(
                "Items not in ascending order. previous: [%s], current: [%s], item: [%s], row_index_r: [%s]"
                % (previous, value, item, col_index_nr))

        if value == item or approximate_match and (
                ln - 1 == row_nr and value < item or previous and value > item > previous):
            return evaluate_spreadsheet(spreadsheet, sheet, search_index, row_nr)
        previous = value
    return ERR_NA

"""
The Excel HLOOKUP function performs a horizontal lookup by searching for a value in the top row of the table and 
returning the value in the same column based on the index_number.
Ref: https://www.techonthenet.com/excel/formulas/hlookup.php
"""
def hlookup(spreadsheet, item, cell_address, row_index_nr, approximate_match=False):
    if not item or item == ERR_NA:
        return ERR_NA
    if approximate_match and not isinstance(item, (float, int)):
        raise Exception("Item must be a number: [%s], row_index_nr: [%s]" % (item, row_index_nr))

    # We assume the range has already been correctly Removed.
    # we can't find it in the cellmap as we changed the cell address to overcome conflict with other cells using
    # the same range
    rng = cell_address
    # rng = get_range(spreadsheet, cell_address)
    item = str(item).lower()
    sheet, start, end = split_range(rng)
    sh, start_col, start_row = split_address(start)
    sh, end_col, end_row = split_address(end)
    start_col_nr, end_col_nr = col2num(start_col), col2num(end_col)
    ln = end_col_nr - start_col_nr + 1
    previous = None
    for col_nr in range(start_col_nr, end_col_nr + 1):
        value = evaluate_spreadsheet(spreadsheet, sheet, num2col(col_nr), start_row)
        if value is None:  # we skip empty cells/columns
            continue

        value = str(value).lower()
        if approximate_match and previous is not None and float(previous) > float(value):

            raise Exception(
                "Items not in ascending order. previous: [%s], current: [%s], item: [%s], row_index_r: [%s]"
                % (previous, value, item, row_index_nr))

        if value == item or approximate_match and (
                ln - 1 == col_nr and value < item or previous is not None and value > item > previous):
            return evaluate_spreadsheet(spreadsheet, sheet, num2col(col_nr), int(start_row) + int(row_index_nr - 1))

        previous = value
    return ERR_NA

# needs unit testing
def linest(*args, **kwargs):
    y = args[0]
    x = args[1]

    if len(args) == 3:
        const = args[2]
        if isinstance(const, str):
            const = (const.lower() == "true")
    else:
        const = True

    degree = kwargs.get('degree', 1)
    # build the vandermonde matrix
    a = np.vander(x, degree + 1)

    if not const:
        # force the intercept to zero
        a[:, -1] = np.zeros((1, len(x)))

    # perform the fit
    (coefs, residuals, rank, sing_vals) = np.linalg.lstsq(a, y)
    return coefs

# needs unit testing
def npv(*args):
    discount_rate = args[0]
    cashflow = args[1]
    return sum([float(x) * (1 + discount_rate) ** -(i + 1) for (i, x) in enumerate(cashflow)])

def match_type_convert(value):
    if type(value) == str:
        value = value.lower()
    elif type(value) == int:
        value = float(value)
    return value

# Ref: https://www.techonthenet.com/excel/formulas/match.php
def match(spreadsheet, item, cell_address, match_type=1):
    if not item or item == ERR_NA:
        return ERR_NA

    previous = None
    result = None
    item = match_type_convert(item)

    def process(col_nr, row_nr, value):
        nonlocal previous, result, match_type, use_col
        if match_type == 1:
            value = match_type_convert(value)
            if previous is not None and value < previous:
                raise Exception(
                    "Array must be in ascending order for match type 1, previous: "
                    "[%s], next: [%s], lookup value: [%s], range: [%s]" % (
                        previous, next, item, rng))
            if item == value:
                result = col_nr if use_col else row_nr
                return False
            if not result:
                result = 0
            elif item < value and previous is not None > previous:
                result -= 1
                return False
        elif match_type == 0:
            if item == match_type_convert(value):
                result = col_nr if use_col else row_nr
                return False

        elif match_type == -1:
            if not result:
                result = 0
            if previous is not None and value > previous:
                raise Exception(
                    "Array must be in descending order for match type 1, previous: [%s], next: [%s], "
                    "lookup_value: [%s, range: [%s]" % (
                        previous, next, item, rng))
            if item == value:
                result = col_nr if use_col else row_nr
                return True
            elif item < value:
                result = col_nr if use_col else row_nr
            elif value > item and previous is not None and item < previous:
                result += 1
                return True
            elif item < value and previous is not None > previous:
                result -= 1
                return False
        else:
            raise Exception('Not a valid match_type, valid match_types are (1, 0 ,-1)')

        previous = value
        return True

    rng = get_range(spreadsheet, cell_address)
    sheet, start, end = split_range(rng)
    sh, start_col, start_row = split_address(start)
    sh, end_col, end_row = split_address(end)
    if start_col == end_col:
        use_col = False
    elif start_row == end_row:
        use_col = True
    else:
        raise Exception("Match only supports a single search row or column. Address: [%s]", rng)
    iterate_cells(spreadsheet, rng, process)
    if result is not None:
        return result - col2num(start_col) + 1 if use_col else result - int(start_row) + 1
    return ERR_NA

# Excel Reference: https://support.office.com/en-us/article/MOD-function-9b6cd169-b6ee-406a-a97b-edf2a9dc24f3
def mod(spreadsheet, nb, q):
    if not isinstance(nb, int):
        raise TypeError("%s is not an integer" % str(nb))
    elif not isinstance(q, int):
        raise TypeError("%s is not an integer" % str(q))
    else:
        return nb % q

# Excel reference: https://support.office.com/en-us/article/COUNT-function-a59cd7fc-b623-4d93-87a4-d23bf411294c
def count(spreadsheet, cell_address):
    counter = 0

    def process(col_nr, row_nr, value):
        nonlocal counter
        if is_nr(value):
            counter += 1
        return True

    # Ensure to use the exact range
    rng = get_range(spreadsheet, cell_address)
    iterate_cells(spreadsheet, rng, process)
    # result = to_values(spreadsheet, rng, True, is_nr);
    return counter

def iferror(spreadsheet, value, value_if_error):
    # list of errors that are used in excel
    return value_if_error if value in ERROR_VALUES else value
"""
Counta function: counts the non empty cells
Ref: https://exceljet.net/excel-functions/excel-counta-function
"""
def counta(spreadsheet, cell_address):
    counter = 0

    def process(col_nr, row_nr, value):
        nonlocal counter
        if value is not None:
            counter += 1
        return True

    # Ensure to use the exact range
    rng = get_range(spreadsheet, cell_address)
    iterate_cells(spreadsheet, rng, process)
    return counter

# Excel reference: https://support.office.com/en-us/article/COUNTIF-function-e0de10c6-f885-4e71-abb4-1f464816df34
def countif(spreadsheet, cell_address, criteria):
    # WARNING:
    # - wildcards not supported
    # - support of strings with >, <, <=, =>, <> not provided
    rng = get_range(spreadsheet, cell_address)

    def is_valid_cell(value):
        return is_valid(value, criteria)

    return len(to_values(spreadsheet, rng, True, is_valid_cell))

# Ref: https://support.office.com/en-us/article/COUNTIFS-function-dda3dc6e-f74e-4aee-88bc-aa8c2a866842
def countifs(*args):
    arg_list = list(args)
    l = len(arg_list)

    if l % 2 != 0:
        raise Exception('excellib.countifs() must have a pair number of arguments, here %d' % l)

    if l >= 2:
        indexes = find_corresponding_index(args[0], args[1])  # find indexes that match first layer of countif

        remaining_ranges = [elem for i, elem in enumerate(arg_list[2:]) if i % 2 == 0]  # get only ranges
        remaining_criteria = [elem for i, elem in enumerate(arg_list[2:]) if i % 2 == 1]  # get only criteria

        filtered_remaining_ranges = []

        for range in remaining_ranges:  # filter items in remaining_ranges that match valid indexes from first countif layer
            filtered_remaining_range = []

            for index, item in enumerate(range):
                if index in indexes:
                    filtered_remaining_range.append(item)

            filtered_remaining_ranges.append(filtered_remaining_range)

        new_tuple = ()
        for index, range in enumerate(
                filtered_remaining_ranges):  # rebuild the tuple that will be the argument of next layer
            new_tuple += (range, remaining_criteria[index])

        return min(countifs(*new_tuple), len(indexes))  # only consider the minimum number across all layer responses

    else:
        return float('inf')

# Ref: https://support.office.com/en-us/article/ROUND-function-c018c5d8-40fb-4053-90b1-b3e7f61a213c
def xround(spreadsheet, number, num_digits=0):
    if not is_number(number) and number in ERROR_VALUES:
        return number
    elif not is_number(number) and isinstance(number, str):
        return ERR_VALUE
    elif not is_number(number):
        raise TypeError("%s is not a number" % str(number))

    if not is_number(num_digits):
        raise TypeError("%s is not a number" % str(num_digits))

    # if num_digits >= 0:  # round to the right side of the point
    # Refs: https://docs.python.org/2/library/functions.html#round
    # https://gist.github.com/ejamesc/cedc886c5f36e2d075c5
    if isinstance(number, (int, float)):
        result = float(Decimal(repr(number)).quantize(Decimal(repr(pow(10, -num_digits))), rounding=ROUND_HALF_UP))
    return result if num_digits > 0 else int(result)

def roundup(spreadsheet, number, num_digits=0):
    if not is_nr(number) or num_digits != 0:
        raise Exception("Incorrect input: nr: [%s], num_digits: [%s]" % (number, num_digits))
    return math.ceil(number)

def rounddown(spreadsheet, number, num_digits=0):
    if not is_nr(number) or num_digits != 0:
        raise Exception("Incorrect input: nr: [%s], num_digits: [%s]" % (number, num_digits))
    return math.trunc(number)

# Excel reference: https://support.office.com/en-us/article/MID-MIDB-functions-d5f9e25c-d7d6-472e-b568-4ecb12433028
def mid(text, start_num, num_chars):
    text = str(text)
    if type(start_num) != int:
        raise TypeError("%s is not an integer" % str(start_num))
    if type(num_chars) != int:
        raise TypeError("%s is not an integer" % str(num_chars))

    if start_num < 1:
        raise ValueError("%s is < 1" % str(start_num))
    if num_chars < 0:
        raise ValueError("%s is < 0" % str(num_chars))

    return text[start_num:num_chars]

# Ref: https://support.office.com/en-us/article/DATE-function-e36c0c8c-4104-49da-ab83-82328b832349
def date(year, month, day):
    if type(year) != int:
        raise TypeError("%s is not an integer" % str(year))

    if type(month) != int:
        raise TypeError("%s is not an integer" % str(month))

    if type(day) != int:
        raise TypeError("%s is not an integer" % str(day))

    if year < 0 or year > 9999:
        raise ValueError("Year must be between 1 and 9999, instead %s" % str(year))

    if year < 1900:
        year = 1900 + year

    year, month, day = normalize_year(year, month, day)  # taking into account negative month and day values
    date_0 = datetime(1900, 1, 1)
    date = datetime(year, month, day)
    result = (datetime(year, month, day) - date_0).days + 2
    if result <= 0:
        raise ArithmeticError("Date result is negative")
    else:
        return result

# Excel reference: https://support.office.com/en-us/article/YEARFRAC-function-3844141e-c76d-4143-82b6-208454ddc6a8
def yearfrac(start_date, end_date, basis=0):
    def actual_nb_days_isda(start, end):  # needed to separate days_in_leap_year from days_not_leap_year
        y1, m1, d1 = start
        y2, m2, d2 = end

        days_in_leap_year = 0
        days_not_in_leap_year = 0
        year_range = list(range(y1, y2 + 1))

        for y in year_range:
            if y == y1 and y == y2:
                nb_days = date(y2, m2, d2) - date(y1, m1, d1)
            elif y == y1:
                nb_days = date(y1 + 1, 1, 1) - date(y1, m1, d1)
            elif y == y2:
                nb_days = date(y2, m2, d2) - date(y2, 1, 1)
            else:
                nb_days = 366 if is_leap_year(y) else 365

            if is_leap_year(y):
                days_in_leap_year += nb_days
            else:
                days_not_in_leap_year += nb_days

        return days_not_in_leap_year, days_in_leap_year

    # http://svn.finmath.net/finmath%20lib/trunk/src/main/java/net/finmath/time/daycount/DayCountConvention_ACT_ACT_YEARFRAC.java
    def actual_nb_days_afb_alter(start, end):
        y1, m1, d1 = start
        y2, m2, d2 = end
        delta = date(*end) - date(*start)

        if delta <= 365:
            if is_leap_year(y1) and is_leap_year(y2):
                denom = 366
            elif is_leap_year(y1) and date(y1, m1, d1) <= date(y1, 2, 29):
                denom = 366
            elif is_leap_year(y2) and date(y2, m2, d2) >= date(y2, 2, 29):
                denom = 366
            else:
                denom = 365
        else:
            year_range = list(range(y1, y2 + 1))
            nb = 0

            for y in year_range:
                nb += 366 if is_leap_year(y) else 365

            denom = nb / len(year_range)

        return delta / denom

    if not is_number(start_date):
        raise TypeError("start_date %s must be a number" % str(start_date))
    if not is_number(end_date):
        raise TypeError("end_date %s must be number" % str(end_date))
    if start_date < 0:
        raise ValueError("start_date %s must be positive" % str(start_date))
    if end_date < 0:
        raise ValueError("end_date %s must be positive" % str(end_date))

    if start_date > end_date:  # switch dates if start_date > end_date
        temp = end_date
        end_date = start_date
        start_date = temp

    y1, m1, d1 = date_from_int(start_date)
    y2, m2, d2 = date_from_int(end_date)

    if basis == 0:  # US 30/360
        d2 = 30 if d2 == 31 and (d1 == 31 or d1 == 30) else min(d2, 31)
        d1 = 30 if d1 == 31 else d1

        count = 360 * (y2 - y1) + 30 * (m2 - m1) + (d2 - d1)
        result = count / 360

    elif basis == 1:  # Actual/actual
        result = actual_nb_days_afb_alter((y1, m1, d1), (y2, m2, d2))

    elif basis == 2:  # Actual/360
        result = (end_date - start_date) / 360

    elif basis == 3:  # Actual/365
        result = (end_date - start_date) / 365

    elif basis == 4:  # Eurobond 30/360
        d2 = 30 if d2 == 31 else d2
        d1 = 30 if d1 == 31 else d1

        count = 360 * (y2 - y1) + 30 * (m2 - m1) + (d2 - d1)
        result = count / 360

    else:
        raise ValueError("%d must be 0, 1, 2, 3 or 4" % basis)

    return result


if __name__ == '__main__':
    pass



# Excel compiler.
# creates nodes from cells puts them in a graph and then puts that graph in a cellmap.
# Has several functions to set and calculate the changed values

import networkx as nx
from networkx.classes.digraph import DiGraph
from .excel_node import *
from .excel_util import *
from .excel_lib import *
from .excel_wrapper import ExcelOpxWrapper as ExcelWrapperImpl
from .excel_wrapper import OpxRange
from .tokenizer import ExcelParser, f_token
logger = logging.getLogger(__name__)

# Set the sheet into a node
def add_node_to_graph(g, n):
    g.add_node(n)
    g.node[n]['sheet'] = n.sheet

    if isinstance(n, Cell):
        g.node[n]['label'] = n.col + str(n.row)
    else:
        # strip the sheet
        g.node[n]['label'] = n.address()[n.address().find('!') + 1:]

class ExcelCompiler(object):

    def __init__(self, filename=None, *args, **kwargs):
        # Load file
        super(ExcelCompiler, self).__init__()
        self.filename = filename
        # Set the filename
        self.excel = ExcelWrapperImpl(filename=filename)
        self.excel.connect()

        self.log = logging.getLogger("decode.{0}".format(self.__class__.__name__))

    def cell2code(self, cell):
        # Generate python code for the given cell
        if cell.formula is not None:
            exp = shunting_yard(cell.formula or str(cell.value))
            if exp is None or len(exp) == 0:
                raise Exception("Incorrect expression: [%s], cell: [%s], "
                                "cell.formula: [%s], cell.value: [%s]"
                                % (exp, cell, cell.formula, cell.value))

            # Correct the ranges found concerning the excel max col and row
            for e in exp:
                if isinstance(e, RangeNode):
                    e.token.tvalue = correct_range_single(self.excel, e.token.tvalue, cell.sheet)

            ast, root = build_ast(exp)
            code = root.emit(ast, context=Context(cell, self.excel))
        elif isinstance(cell.value, str) and re.match('[A-Za-z]+', cell.value):
            ast = None
            code = cell.value
        else:
            ast = None
            code = str('"' + cell.value + '"' if isinstance(cell.value, str) else cell.value)

        return code, ast

    def gen_graph(self, rng, sheet=None):
        return self.gen_graph_multi({sheet: rng})

    """
    Example usage:
        gen_graph({'Output1': 'A1:B5', 'Output2': 'A1:E8'})
        gen_graph({'Output1': '[A1:B5, AB4:FE85]', 'Output2': 'A1:E8'})
    """

    def gen_graph_multi(self, ranges):
        # When given a starting point (e.g., A6 or A3:B7) on a particular sheet,
        # generate a Spreadsheet instance that captures the logic and control flow of the equations

        # starting points
        # cursheet = sheet if sheet else self.excel.get_active_sheet()
        # self.excel.set_sheet(cursheet)

        # no need to output NumberRows and NumberColumns here,
        # since seed can be a list of unlinked cells
        seeds = None
        try:
            seeds = Cell.make_cells_multi(self.excel, ranges)
            if logger.isEnabledFor(logging.DEBUG):
                logger.info("Ranges [%s] expanded into [%s] cell_len", ranges, seeds)
            return self.gen_graph_intern(seeds)
        except Exception as ex:
            raise Exception("Error making cells, ranges: [%s], "
                            "excel: [%s]" % (ranges, self.excel)) from ex

    def gen_graph_intern(self, seeds):
        # only keep seeds with formulas or numbers
        # Scan seeds for data_types int, float or strings
        seeds = [s for s in seeds if s.formula or
                 isinstance(s.value, (int, float, str))]

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("Filtered seeds (keeping formulas, "
                         "strings, ints, floats): [%s] ", len(seeds))

        todo = [s for s in seeds if s.formula]

        if logger.isEnabledFor(logging.DEBUG):
            logger.info("Todo cells length: [%s]", len(todo))

        # map of all cells
        cellmap = dict([(x.address(), x) for x in seeds])

        # directed graph
        graph = nx.DiGraph()

        # match the info in cellmap
        for c in list(cellmap.values()):
            add_node_to_graph(graph, c)

        cursheet = self.excel.get_active_sheet()
        while todo:
            c1 = todo.pop()
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Processing formula cell: [%s]", c1)

            # set the current sheet so relative addresses resolve properly
            if c1.sheet != cursheet:
                cursheet = c1.sheet
                self.excel.set_sheet(cursheet)

            try:
                pystr, ast = self.cell2code(c1)  # parse the formula into code
                c1.python_expression = pystr  # set the code & compile it (will flag problems sooner rather than later)
                c1.compile()
            except Exception as ex:
                raise Exception(
                    "Error compiling/cell2code cell: [%s], todo length: [%s]" % (c1, len(todo))) from ex

            if ast is not None:
                # get all the cells/ranges this formula refers to
                # No need to correct for '$' as it has been moved to the cell2node() function
                deps = [x.tvalue for x in ast.nodes() if isinstance(x, RangeNode)]
                deps = uniqueify(deps)  # remove dupes
                # deps = [x.tvalue.replace('$', '') for x in ast.nodes() if isinstance(x, RangeNode)]
                # deps = uniqueify(deps)  # remove dupes

                for dep in deps:
                    # Ensure to always set the active sheet, especially in case no dependency sheet is found,
                    # that means the dependency occurs on the same sheet as cursheet,
                    # but in case we just had a dependency to another
                    # sheet that can result in using the wrong sheet if not set.
                    dep_sheet = create_sheet(dep)
                    dep_sheet = dep_sheet if dep_sheet else cursheet
                    self.excel.set_sheet(dep_sheet if dep_sheet else cursheet)

                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug("Processing dependency [%s!%s] of formula cell: [%s]", dep_sheet, dep, c1)

                    # if the dependency is a multi-cell range, create a range object
                    rng_address, rng = None, None
                    if is_range(dep):
                        try:
                            cells, nrows, ncols, rng_details = Cell.make_cells(self.excel, dep, sheet=dep_sheet,
                                                                               parent_ast=ast, cellmap=cellmap)

                            # this will make sure we always have an absolute address
                            rng_address = create_address(rng_details, sheet=dep_sheet)

                            # No cells created if not needed (already exists in cell map)
                            if cells is None or rng_address in cellmap:
                                # already dealt with this range
                                # add an edge from the range to the parent
                                graph.add_edge(cellmap[rng_address], cellmap[c1.address()])

                                if logger.isEnabledFor(logging.DEBUG):
                                    logger.debug(
                                        "Skipping range dependency [%s] of dependency: [%s], "
                                        "of formula cell: [%s], already present",
                                        rng_address, dep, c1)
                                continue

                            # update the original range as the returned one is probably more specific if it wasn't a list
                            rng = CellRange(rng_details, dep, None, sheet=dep_sheet)
                            # cells, nrows, ncols = Cell.make_cells(self.excel, {cursheet: dep})

                            # save the range
                            # Ensure to use the original address to store it as other need it to find it.
                            cellmap[rng_address] = rng
                            # cellmap[rng.address()] = rng
                            # add an edge from the range to the parent
                            add_node_to_graph(graph, rng)
                            graph.add_edge(rng, cellmap[c1.address()])
                            # cells in the range should point to the range as their parent
                            target = rng
                        except Exception as ex:
                            raise Exception(
                                "Error making cells, cell: [%s], ast: [%s], dep: [%s], "
                                "rng address: [%s], rng: [%s], todo_len: [%s], seeds_len: [%s]" %
                                (c1, ast, dep, rng_address, rng, len(todo), len(seeds))) from ex
                    else:
                        # not a range, create the cell object
                        cells = (Cell.resolve_cell(self.excel, dep, sheet=dep_sheet),)
                        target = cellmap[c1.address()]

                    # process each cell
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug("Adding [%s] row-cells for dependency [%s!%s]", len(cells), dep_sheet, dep)

                    for c2 in flatten(cells):
                        # Check that cells are None
                        if c2 is not None:
                            # if we haven't treated this cell already
                            if c2.address() in cellmap:
                                if logger.isEnabledFor(LOG_FINE):
                                    logger.log(LOG_FINE, "Skipping [%s] cell for dependency [%s!%s], "
                                                         "already present", c2,
                                               dep_sheet, dep)
                            else:
                                if c2.formula:
                                    # cell with a formula, needs to be added to the todo list
                                    todo.append(c2)
                                else:
                                    try:
                                        # constant cell, no need for further processing, just remember to set the code
                                        pystr, ast = self.cell2code(c2)
                                        c2.python_expression = pystr
                                        c2.compile()
                                    except Exception as ex:
                                        raise Exception(
                                            "Error compiling/cell2code dependency cell, cell dep: [%s] cell: [%s], "
                                            "ast: [%s], dep: [%s], "
                                            "todo: [%s], seeds: [%s]" % (c2, c1, ast, dep, todo, seeds)) from ex

                                # save in the cellmap
                                if logger.isEnabledFor(LOG_FINE):
                                    logger.log(LOG_FINE, "Adding cell [%s] to cellmap for dependency [%s!%s]", c2,
                                               dep_sheet, dep)
                                cellmap[c2.address()] = c2
                                # add to the graph
                                add_node_to_graph(graph, c2)

                            # add an edge from the cell to the parent (range or cell)
                            graph.add_edge(cellmap[c2.address()], target)

        if logger.isEnabledFor(logging.INFO):
            logger.info("Graph construction done, %s nodes, %s edges, %s cellmap entries",
                        len(graph.nodes()), len(graph.edges()), len(cellmap))
        sp = Spreadsheet(graph, cellmap)
        return sp

def shunting_yard(expression):
    """
    Tokenize an excel formula expression into reverse polish notation

    Core algorithm taken from wikipedia with varargs extensions from
    http://www.kallisti.net.nz/blog/2008/02/extension-to-the-shunting-yard-algorithm-to-allow-variable-numbers-of-arguments-to-functions/
    """
    # remove leading =

    if expression.startswith('='):
        expression = expression[1:]

    p = ExcelParser()
    p.parse(expression)

    # insert tokens for '(' and ')', to make things clearer below
    tokens = []
    for t in p.tokens.items:
        if t.ttype == "function" and t.tsubtype == "start":
            t.tsubtype = ""
            tokens.append(t)
            tokens.append(f_token('(', 'arglist', 'start'))
        elif t.ttype == "function" and t.tsubtype == "stop":
            tokens.append(f_token(')', 'arglist', 'stop'))
        elif t.ttype == "subexpression" and t.tsubtype == "start":
            t.tvalue = '('
            tokens.append(t)
        elif t.ttype == "subexpression" and t.tsubtype == "stop":
            t.tvalue = ')'
            tokens.append(t)
        else:
            tokens.append(t)

    # http://office.microsoft.com/en-us/excel-help/calculation-operators-and-precedence-HP010078886.aspx
    operators = {':': Operator(':', 8, 'left'), '': Operator(' ', 8, 'left'), ',': Operator(',', 8, 'left'),
                 'u-': Operator('u-', 7, 'left'), '%': Operator('%', 6, 'left'), '^': Operator('^', 5, 'left'),
                 '*': Operator('*', 4, 'left'), '/': Operator('/', 4, 'left'), '+': Operator('+', 3, 'left'),
                 '-': Operator('-', 3, 'left'), '&': Operator('&', 2, 'left'), '=': Operator('=', 1, 'left'),
                 '<': Operator('<', 1, 'left'), '>': Operator('>', 1, 'left'), '<=': Operator('<=', 1, 'left'),
                 '>=': Operator('>=', 1, 'left'), '<>': Operator('<>', 1, 'left')}

    output = collections.deque()
    stack = []
    were_values = []
    arg_count = []

    for t in tokens:
        if t.ttype == "operand":

            output.append(create_node(t))
            if were_values:
                were_values.pop()
                were_values.append(True)

        elif t.ttype == "function":

            stack.append(t)
            arg_count.append(0)
            if were_values:
                were_values.pop()
                were_values.append(True)
            were_values.append(False)

        elif t.ttype == "argument":

            while stack and (stack[-1].tsubtype != "start"):
                output.append(create_node(stack.pop()))

            if were_values.pop():
                arg_count[-1] += 1
            were_values.append(False)

            if not len(stack):
                raise Exception("Mismatched or misplaced parentheses")

        elif t.ttype.startswith('operator'):

            if len(stack) > 0 and t.ttype.endswith('-prefix') and t.tvalue == "-":
                o1 = operators['u-']
            else:
                o1 = operators[t.tvalue]

            while stack and stack[-1].ttype.startswith('operator'):
                if stack[-1].ttype.endswith('-prefix') and stack[-1].tvalue == "-":
                    o2 = operators['u-']
                else:
                    o2 = operators[stack[-1].tvalue]

                if ((o1.associativity == "left" and o1.precedence <= o2.precedence)
                        or
                        (o1.associativity == "right" and o1.precedence < o2.precedence)):

                    output.append(create_node(stack.pop()))
                else:
                    break

            stack.append(t)

        elif t.tsubtype == "start":
            stack.append(t)

        elif t.tsubtype == "stop":

            while stack and stack[-1].tsubtype != "start":
                output.append(create_node(stack.pop()))

            if not stack:
                raise Exception("Mismatched or misplaced parentheses")

            stack.pop()

            if stack and stack[-1].ttype == "function":
                f = create_node(stack.pop())
                a = arg_count.pop()
                w = were_values.pop()
                if w:
                    a += 1
                f.num_args = a
                output.append(f)

    while stack:
        if stack[-1].tsubtype == "start" or stack[-1].tsubtype == "stop":
            raise Exception("Mismatched or misplaced parentheses")

        output.append(create_node(stack.pop()))

    # convert to list
    result = [x for x in output]
    return result

def build_ast(expression):
    """build an AST from an Excel formula expression in reverse polish notation"""

    # use a directed graph to store the tree
    graph = DiGraph()

    stack = []

    for n in expression:
        # Since the graph does not maintain the order of adding nodes/edges
        # add an extra attribute 'pos' so we can always sort to the correct order
        if isinstance(n, OperatorNode):
            if n.ttype == "operator-infix":
                arg2 = stack.pop()
                arg1 = stack.pop()
                graph.add_node(arg1, pos=1)
                graph.add_node(arg2, pos=2)
                graph.add_edge(arg1, n)
                graph.add_edge(arg2, n)
            else:
                arg1 = stack.pop()
                graph.add_node(arg1, pos=1)
                graph.add_edge(arg1, n)
        elif isinstance(n, FunctionNode):
            args = [stack.pop() for _ in range(n.num_args)]
            args.reverse()
            for i, a in enumerate(args):
                graph.add_node(a, pos=i)
                graph.add_edge(a, n)
                # for i in range(n.num_args):
                #    G.add_edge(stack.pop(),n)
        else:
            graph.add_node(n, pos=0)

        stack.append(n)

    if len(stack) < 1:
        raise Exception("Error building ast, expression: [%s], graph: [%s]" % (expression, graph))
    return graph, stack.pop()

class Context(object):
    """A small context object that nodes in the AST can use to emit code"""

    def __init__(self, curcell, excel):
        # the current cell for which we are generating code
        self.curcell = curcell
        # a handle to an excel instance
        self.excel = excel

def create_node(t):
    """Simple factory function"""
    if t.ttype == "operand":
        if t.tsubtype == "range":
            return RangeNode(t)
        else:
            return OperandNode(t)
    elif t.ttype == "function":
        return FunctionNode(t)
    elif t.ttype.startswith("operator"):
        return OperatorNode(t)
    else:
        return ASTNode(t)

class Operator:
    """Small wrapper class to manage operators during shunting yard"""

    def __init__(self, value, precedence, associativity):
        self.value = value
        self.precedence = precedence
        self.associativity = associativity

class Spreadsheet(object):

    def __init__(self, g, cellmap):
        super(Spreadsheet, self).__init__()
        self.graph = g
        self.cellmap = cellmap
        self.__call_stack = []
        self.params = None

    def contains(self, cell):
        return cell in self.cellmap

    def plot_graph(self):
        import matplotlib.pyplot as plt

        pos = nx.spring_layout(self.graph, iterations=2000)
        nx.draw_networkx_nodes(self.graph, pos)
        nx.draw_networkx_edges(self.graph, pos, arrows=True)
        nx.draw_networkx_labels(self.graph, pos)
        plt.show()

    def set_value(self, cell, val, is_addr=True):
        if is_addr is not None:
            cell = self.cellmap[cell]
            if cell.value != val:
                # We should always reset the successors when the value is changes,
                # also when it's changed from a None value
                # to a not-None value, else the others aren't informed, and will not reset it's value such that the old
                # value will be used when evaluated.

                self.reset_all(cell, True)
                cell.value = val
                # Because we set a value we indicate it as not-dirty. Suppose the cell contains a formula, like a
                # reference to another cell, setting the value will overwritte the behavior at that moment.
                # When we "read"
                # the value through evaluation, it would (if not indicating it as not-dirty) evaluate the formula as
                # it's dirty through the reset call in  this method, such that it will set/store the value of the
                # reference and overwrite the set value "manually" by calling this method (as if it wasn't set)
                # Note: till it will be reset, such that it becomes dirty,
                # the set value will be returned. When it's dirty
                # will be be evaluated that will result in using the referenced value (again)
                cell.dirty = False

    def reset(self, cell):
        self.reset_all(cell, not cell.dirty or isinstance(cell, CellRange))
        # self.reset_all(cell, cell.value is not None or isinstance(cell, CellRange))

    def reset_all(self, cell, reset_successors):
        if logger.isEnabledFor(LOG_FINE):
            logger.log(LOG_FINE, "Resetting cell, dirty: [%s], cell: [%s]", cell.dirty, cell)
        if cell.reset_inprogress:
            return  # Already been reset, let's skip it.
            # raise Exception("Cell reset in progress: [%s]" % (cell))
        try:
            cell.reset_inprogress = True
            # Note: we split the reset method,
            # as before it would only perform the reset in case the cell value isn't None
            # already. However, that was restricted
            # in case it was already None, such that a change in value would not reset
            # the successors such that the successors would use the old value.
            # CellRange instances are skipped as their value isn't used
            if isinstance(cell, Cell):
                cell.value = None

            # If already dirty, no need to continue, we assume that it's successors been made dirty as well.
            # The value of CellRange isn't used so we have to skip it
            if not cell.dirty or isinstance(cell, CellRange):
                if isinstance(cell, Cell):
                    cell.dirty = True
                if reset_successors:
                    if logger.isEnabledFor(LOG_FINE):
                        childs = '\n'.join(str(p) for p in self.graph.succ[cell].keys())
                        logger.log(LOG_FINE, "Resetting children of: [%s]:\n[%s]", cell.address(), childs)
                    list(map(self.reset, self.graph.successors(cell)))
        finally:
            cell.reset_inprogress = False

    def print_value_tree(self, addr, indent):
        cell = self.cellmap[addr]
        print(("%s %s = %s" % (" " * indent, addr, cell.value)))
        for c in self.graph.predecessors_iter(cell):
            self.print_value_tree(c.address(), indent + 1)

    # def recalculate(self, addr):
    #     for c in list(self.cellmap.values()):
    #         if isinstance(c, CellRange):
    #             self.evaluate_range(c, is_addr=False)
    #         else:
    #             self.evaluate(c, is_addr=False)
    #         return self.evaluate(c)

    # def evaluate_range(self, rng, is_addr=True):
    #     if is_addr:
    #         rng = self.cellmap[rng]
    #
    #     # its important that [] gets treated as false here
    #     if rng.value:
    #         return rng.value
    #     elif isinstance(rng.value, str):
    #         return rng.value
    #
    #     cells, nrows, ncols = rng.celladdress, rng.nrows, rng.ncols
    #     if 1 == nrows or 1 == ncols:
    #         data = [self.evaluate(c) for c in cells]
    #     else:
    #         data = [[self.evaluate(c) for c in cells[i]] for i in range(len(cells))]
    #
    #     # Ensure to indicate the CellRange instance isn't dirty anymore, else it will get ignored next time such that
    #     # it's successors aren't reset and return the wrong value (example: B1(Cell)->
    #     B1:B2(CellRange)-> Vlookup(Cell)
    #     rng.value = data
    #     rng.dirty = False
    #     return data

    def evaluate(self, cell, is_addr=True):
        if is_addr is not None:
            cell = self.cellmap[cell]
            if cell.eval_inprogress:
                raise Exception("Evaluation of cell already in progress, cell: [%s], call stack:\n%s"
                                % (cell, self.call_stack_string(self.__call_stack)))

            # We only evaluate cells that have a formula and are dirty. In case they aren't dirty we return their
            # current value. We don't check if the value isn't None, as None might be a valid result of a formula
            # which can cause infinite, or unnecessary loops

            if not cell.formula or not cell.dirty:
                return cell.value
                # if not cell.formula:
                #     return cell.value

        def eval_cell(address):
            # return address
            try:
                # Note: we have to evaluate the cell and return the value, else an expression like "A3+A4" will fail
                # as python will try to "add" strings. This expression is processed by python directly and not through
                # a function in the excel_lib
                value = self.evaluate(address)
                if not value:
                    try:
                        if value is None or len(value) == 0 and cell.formula:
                            # The split will fail in case it's not a valid cell address
                            split_address(cell.formula[1:])
                            # In case it concerns a cell reference,
                            # it should return 0 in case the references cell is None
                            value = 0
                    except Exception:
                        pass
                return value
            # return self.evaluate(address)
            except:
                raise self.create_exception("Problem evaluating cell [%s]", address)

        def eval_range(rng):
            try:
                # We return the range and let the "user" (function) decide when to evaluate.
                # Example: in case of a (v/h)lookup function we don't want to evaluate the complete search
                # range before feeding it to the lookup function, as the lookup function only use small parts
                # of the search rang. And evaluating a huge range can easily lead to circular references.
                return rng
                # return self.evaluate_range(rng)
            except:
                raise self.create_exception("Problem evaluating cell [%s]", rng)

        try:
            cell.eval_inprogress = True
            if logger.isEnabledFor(LOG_FINE):
                logger.log(LOG_FINE, "Evaluating: %s, %s", cell.address(), cell.python_expression)

            if is_eval_allowed(self, cell):
                # indicate it's not dirty anymore before evaluation occurs to overcome infinite recursive loops.
                cell.dirty = False

                self.__call_stack.append(str(cell))
                vv = eval(cell.compiled_expression)
            else:
                vv = cell.compiled_expression

            cell.dirty = False
            cell.value = vv
            digits = get_zeros(cell)

            if digits is not None and not isinstance(cell.value, str) and cell.value is not None:
                cell.value = round(cell.value, digits)

        except Exception as ex:
            if str(ex).startswith("Problem evaluating"):
                raise ex
            else:
                raise Exception(
                    "Problem evaluating: [%s], call stack: \n%s" % (
                        cell, self.call_stack_string(self.__call_stack))) from ex
        finally:
            cell.eval_inprogress = False
            self.__call_stack.pop()
            if len(self.__call_stack) > 10000:
                self.__call_stack = []
                logger.error("Call stack has reached maximum. Cell: [%s]", cell)

        return cell.value

    def create_exception(self, msg, rng):
        if rng in self.cellmap:
            rng = self.cellmap[rng]
        return Exception(msg % rng)

    def call_stack_string(self, call_stack):
        if len(call_stack) > 0:
            result = ''
            for x in call_stack:
                result += x + '\n'
            return result
        return ''



"""
Created on 16 mrt. 2017

@author: Matthijs Kruger
"""
from converter import excel_lib
from .excel_util import is_range, split_range, split_address, get_linest_degree

class ASTNode(object):
    """A generic node in the AST"""

    def __init__(self, token):
        super(ASTNode, self).__init__()
        self.token = token

    def __str__(self):
        return self.token.tvalue

    def __getattr__(self, name):
        return getattr(self.token, name)

    def children(self, ast):
        args = ast.predecessors(self)
        args = sorted(args, key=lambda x: ast.node[x]['pos'])
        # args.reverse()
        return args

    def parent(self, ast):
        args = ast.successors(self)
        return args if args else None

    def emit(self, ast, context=None):
        """Emit code"""
        # self.token.tvalue

class OperatorNode(ASTNode):
    def __init__(self, *args):
        super(OperatorNode, self).__init__(*args)

        # convert the operator to python equivalents
        self.opmap = {
            "^": "**",
            "=": "==",
            "&": "+",
            "": "+",  # union
            "<>": "!=",  # Added to support the <> (not equals), as <> doesn't seem to work and != does
        }

    def emit(self, ast, context=None):
        xop = self.tvalue
        args = self.children(ast)  # Get the arguments
        op = self.opmap.get(xop, xop)

        if self.ttype == "operator-prefix":
            return "-" + args[0].emit(ast, context=context)

        parent = self.parent(ast)
        # don't render the ^{1,2,..} part in a linest formula
        # TODO: bit of a hack
        ss = None
        if op == "**":
            if parent is not None and parent == "linest":
                return emit_arg(args, ast, context, 0)

        if op == "<" or op == "<=":
            # aa = emit_arg(args, ast, context, 0)
            # ss = "(" + aa + " if " + aa + " is not None else float('inf'))" + op + emit_arg(args, ast, context, 1)
            ss = surround('less_then', emit_arg(args, ast, context, 0), emit_arg(args, ast, context, 1),
                          str(op == '<='))
        elif op == ">" or op == ">=":
            # aa = emit_arg(args, ast, context, 1)
            # ss = emit_arg(args, ast, context, 0) + op + "(" + aa + " if " + aa + " is not None else float('inf'))"
            ss = surround('greater_then', emit_arg(args, ast, context, 0), emit_arg(args, ast, context, 1),
                          str(op == '>='))
        elif xop == '&' and op == '+':
            ss = to_safe_str(emit_arg(args, ast, context, 0)) + op + to_safe_str(emit_arg(args, ast, context, 1))
            # ss = 'safe_str(' + emit_arg(args, ast, context, 0) + ')' + op +
            # 'safe_str(' + emit_arg(args, ast, context, 1) + ')'
        elif op == '*':
            ss = surround('multiply', emit_arg(args, ast, context, 0), emit_arg(args, ast, context, 1))
        elif op == '/':
            ss = surround('divide', emit_arg(args, ast, context, 0), emit_arg(args, ast, context, 1))
        elif op == '+':
            ss = surround('plus', emit_arg(args, ast, context, 0), emit_arg(args, ast, context, 1))
        elif op == '-':
            ss = surround('minus', emit_arg(args, ast, context, 0), emit_arg(args, ast, context, 1))
        else:
            ss = emit_arg(args, ast, context, 0) + op + emit_arg(args, ast, context, 1)

        # avoid needless parentheses
        if ss and parent and not isinstance(parent, FunctionNode):
            ss = "(" + ss + ")"

        return ss

def emit_arg(args, ast, context, index):
    return args[index].emit(ast, context=context)

def surround(fn_name, *args):
    return fn_name + '(' + ",".join([n for n in args]) + ')'

def to_safe_str(value):
    return surround("safe_str", value)

class OperandNode(ASTNode):
    def __init__(self, *args):
        super(OperandNode, self).__init__(*args)

    def emit(self, ast, context=None):
        t = self.tsubtype

        if t == "logical":
            return str(self.tvalue.lower() == "true")
        elif t == "text" or t == "error":
            # if the string contains quotes, escape them
            val = self.tvalue.replace('"', '\\"')
            return '"' + val + '"'
        else:
            return str(self.tvalue)

class RangeNode(OperandNode):
    """Represents a spreadsheet cell or range, e.g., A5 or B3:C20"""
    def __init__(self, *args):
        super(RangeNode, self).__init__(*args)

    def emit(self, ast, context=None):
        # resolve the range into cells
        rng = self.tvalue.replace('$', '')
        sheet = context.curcell.sheet + "!" if context else ""
        if is_range(rng):
            sh, start, end = split_range(rng)
            if sh is not None:
                ss = 'eval_range("' + rng + '")'
            else:
                ss = 'eval_range("' + sheet + rng + '")'
        else:
            sh, col, row = split_address(rng)
            if sh is not None:
                ss = 'eval_cell("' + rng + '")'
            else:
                ss = 'eval_cell("' + sheet + rng + '")'
        return ss

class FunctionNode(ASTNode):
    """AST node representing a function call"""
    def __init__(self, *args):
        super(FunctionNode, self).__init__(*args)
        self.numargs = 0

        # map excel functions onto their python equivalents
        self.funmap = excel_lib.FUNCTION_MAP

    def emit(self, ast, context=None):
        fun = self.tvalue.lower()
        expr = ''

        # Get the arguments
        args = None
        if len(ast.pred) > 0:
            args = self.children(ast)

        if fun == "atan2":
            # swap arguments
            expr = "atan2(%s,%s)" % (emit_arg(args, ast, context, 1), emit_arg(args, ast, context, 0))
        elif fun == "pi":
            # constant, no parents
            expr = "pi"
        elif fun == "sin":
            expr = "sin(%s,%s)" % (emit_arg(args, ast, context, 0), emit_arg(args, ast, context, 0))
        elif fun == "if":
            # inline the if
            if len(args) == 2:
                expr = "%s if %s else 0" % (emit_arg(args, ast, context, 1), emit_arg(args, ast, context, 0))
            elif len(args) == 3:
                expr = "(%s if %s else %s)" % (emit_arg(args, ast, context, 1), emit_arg(args, ast, context, 0),
                                               emit_arg(args, ast, context, 2))
            else:
                raise Exception("if with %s arguments not supported" % len(args))

        elif fun == "array":
            expr += '['
            if len(args) == 1:
                # only one row
                expr += emit_arg(args, ast, context, 0)
            else:
                # multiple rows
                expr += ",".join(['[' + n.emit(ast, context=context) + ']' for n in args])

            expr += ']'
        elif fun == "arrayrow":
            # simply create a list
            expr += ",".join([n.emit(ast, context=context) for n in args])
        elif fun == "linest" or fun == "linestmario":

            expr = fun + "(" + ",".join([n.emit(ast, context=context) for n in args])

            if not context:
                degree, coef = -1, -1
            else:
                # linests are often used as part of an array formula spanning multiple cells,
                # one cell for each coefficient.  We have to figure out where we currently are
                # in that range
                degree, coef = get_linest_degree(context.excel, context.curcell)

            # if we are the only linest (degree is one) and linest is nested -> return vector
            # else return the coef.
            if degree == 1 and self.parent(ast):
                if fun == "linest":
                    expr += ",degree=%s)" % degree
                else:
                    expr += ")"
            else:
                if fun == "linest":
                    expr += ",degree=%s)[%s]" % (degree, coef - 1)
                else:
                    expr += ")[%s]" % (coef - 1)

        elif fun == "and":
            expr = "all([" + ",".join([n.emit(ast, context=context) for n in args]) + "])"
        elif fun == "or":
            expr = "any([" + ",".join([n.emit(ast, context=context) for n in args]) + "])"
        elif fun == "not":
            if len(args) > 1: raise Exception("Too many arguments for NOT function (1 expected): [%s], args: [%s]",
                                              len(args), args)
            expr = "not(" + ",".join([n.emit(ast, context=context) for n in args]) + ")"
        elif fun == "row":
            f = self.funmap.get(fun, fun)
            if args is None or len(args) == 0:
                sh, ad = context.curcell.address().split('!')
                args = [ad]
                expr = f + '("' + ",".join([n.replace('"', '\\"') for n in args]) + '")'
            else:
                expr = f + '("' + ",".join([n.tvalue.replace('"', '\\"') for n in args]) + '")'
        else:
            # map to the correct name
            f = self.funmap.get(fun, fun)
            expr = f + "(self, " + ",".join([n.emit(ast, context=context) for n in args]) + ")"
        return expr
